/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Dictionary facility.
 *
 * Handle the dictionary, that contains definitions for all Diameter objects supported by the application.
 *
 * See dictionary-api.h and dictionary.h for more information on the functions 
 *
 */

#include <pthread.h>
#include <errno.h>
#include <string.h>
#include <assert.h>

#include "waaad-internal.h"
#include "dict-base.h"

/*******************************************************************************************************/
/*******************************************************************************************************/
/*                                                                                                     */
/*                                  Types and global objects                                           */
/*                                                                                                     */
/*******************************************************************************************************/
/*******************************************************************************************************/

/* Names of the base types */
char * type_base_name[] = { "GROUPED", "OCTETSTRING", "INTEGER32", "INTEGER64", "UNSIGNED32", "UNSIGNED64", "FLOAT32", "FLOAT64"};

/* The read-write lock for the whole dictionary */
static pthread_rwlock_t dict_lock;

/* forward declaration */
struct __dict_object;

/* List entities */
typedef struct __dict_list {
	struct __dict_list	*next;	/* The next object in the list */
	struct __dict_list	*prev;	/* The previous object in the list */
	struct __dict_object	*o;	/* pointer to the start of this object (may use "offsetof" instead) */
} _dict_list_t;

/* The number of lists in an object */
#define NB_LISTS_PER_OBJ	3

/* This is the internal description of a dictionary object */
typedef struct __dict_object {
	dict_object_type_t	type;	/* What type of object is this? */
	int			objeyec;/* eyecatcher for this object */
	int			typeyec;/* eyecatcher for this type of object */
	
	union {
		dict_vendor_data_t	vendor;
		dict_application_data_t	application;
		dict_type_data_t	type;
		dict_type_enum_data_t	type_enum;
		dict_avp_data_t		avp;
		dict_cmd_data_t		cmd;
		dict_rule_data_t	rule;
	} data;				/* The data of this object */
	
	struct __dict_object *	parent; /* The parent of this object, if any */
	
	_dict_list_t		list[NB_LISTS_PER_OBJ];/* used to chain objects.*/
	/* More information about the lists :
	
	 - the use for each list depends on the type of object. See detail bellow.
	 
	 - a sentinel for a list does not have its 'o' field set. (this is the criteria to detect end of a loop)
	 
	 - The lists are always ordered. The criteria are described bellow. the functions to order them are referenced in dict_obj_info
	 
	 - The dict_lock must be held for any list operation.
	 
	 => VENDORS:
	 list[0]: list of the vendors, ordered by their id. The sentinel is g_dict_vendors (vendor with id 0)
	 list[1]: sentinel for the list of AVPs from this vendor, ordered by AVP code.
	 list[2]: sentinel for the list of AVPs from this vendor, ordered by AVP name.
	 
	 => APPLICATIONS:
	 list[0]: list of the applications, ordered by their id. The sentinel is g_dict_applications (application with id 0)
	 list[1]: not used
	 list[2]: not used.
	 
	 => TYPES:
	 list[0]: list of the types, ordered by their names. The sentinel is g_list_types.
	 list[1]: sentinel for the type_enum list of this type, ordered by their constant name.
	 list[2]: sentinel for the type_enum list of this type, ordered by their constant value.
	 
	 => TYPE_ENUMS:
	 list[0]: list of the contants for a given type, ordered by the constant name. Sentinel is a (list[1]) element of a TYPE object.
	 list[1]: list of the contants for a given type, ordered by the constant value. Sentinel is a (list[2]) element of a TYPE object.
	 list[2]: not used
	 
	 => AVPS:
	 list[0]: list of the AVP from a given vendor, ordered by avp code. Sentinel is a list[1] element of a VENDOR object.
	 list[1]: list of the AVP from a given vendor, ordered by avp name. Sentinel is a list[2] element of a VENDOR object.
	 list[2]: sentinel for the rule list that apply to this AVP.
	 
	 => COMMANDS:
	 list[0]: list of the commands, ordered by their names. The sentinel is g_list_cmd_name.
	 list[1]: list of the commands, ordered by their command code and 'R' flag. The sentinel is g_list_cmd_code.
	 list[2]: sentinel for the rule list that apply to this command.
	 
	 => RULES:
	 list[0]: list of the rules for a given (grouped) AVP or Command, ordered by the AVP name to which they refer. sentinel is list[2] of a command or (grouped) avp.
	 list[1]: not used
	 list[2]: not used.
	 
	 */
	 
	 /* (later: add the sentinel for registered callbacks) */
	
} _dict_object_t;

/* Cast opaque type to internal type */
#define _O( object ) ((_dict_object_t *) (object))

#define OBJECT_EYECATCHER	(0x0b13c7)

/* Sentinels for the different lists */
static _dict_object_t	g_dict_vendors;	
static _dict_object_t	g_dict_applications;
static _dict_list_t	g_list_types;
static _dict_list_t	g_list_cmd_name;
static _dict_list_t	g_list_cmd_code;

/* Special command object for answers with the 'E' bit set */
static _dict_object_t   g_dict_cmd_error;

/* Functions to dump the data structure */
typedef void (*dump_data_t) 	  ( void * data );

/* Forward declarations of such functions */
static void dump_vendor_data 	  ( void * data );
static void dump_application_data ( void * data );
static void dump_type_data 	  ( void * data );
  /* the dump function for enum has a different prototype since it need the datatype */
static void dump_avp_data 	  ( void * data );
static void dump_command_data 	  ( void * data );
static void dump_rule_data 	  ( void * data );

/* Functions to search an element in the dictionary */
typedef int (*search_fct_t) 	( int criteria, void * what, _dict_object_t **result );

/* Forward declarations of such functions */
static int search_vendor 	( int criteria, void * what, _dict_object_t **result );
static int search_application   ( int criteria, void * what, _dict_object_t **result );
static int search_type 		( int criteria, void * what, _dict_object_t **result );
static int search_type_enum 	( int criteria, void * what, _dict_object_t **result );
static int search_avp		( int criteria, void * what, _dict_object_t **result );
static int search_cmd		( int criteria, void * what, _dict_object_t **result );
static int search_rule		( int criteria, void * what, _dict_object_t **result );


static struct _dict_obj_info {
	dict_object_type_t 	type; 		/* information for this type */
	char *			name;		/* string describing this object, for debug */
	size_t			datasize;	/* The size of the data structure */
	int			parent;		/* 0: never; 1: may; 2: must */
	dict_object_type_t	parenttype;	/* The type of the parent, when relevant */
	int			eyecatcher;	/* A kind of signature for this object */
	dump_data_t		dump_data;	/* The function to dump the data section */
	search_fct_t		search_fct;	/* The function to search an object of this type */
	int			haslist[NB_LISTS_PER_OBJ];	/* Tell if this list is used */
	int			count;		/* The number of objects of this type in the dictionary */
} dict_obj_info[] = { { 0, "(error)", 0, 0, 0, 0, NULL, NULL, {0, 0, 0}, 0 }

	/* type			 name		datasize		     parent  	parenttype 
			eyecatcher		dump_data	  	search_fct,		haslist[] 	count*/

	,{ DICT_VENDOR,		"VENDOR",	sizeof(dict_vendor_data_t),	0, 	0,
			OBJECT_EYECATCHER + 1, 	dump_vendor_data, 	search_vendor, 		{ 1, 0, 0 },	0 }
	
	,{ DICT_APPLICATION,	"APPLICATION",	sizeof(dict_application_data_t),1, 	DICT_VENDOR,
			OBJECT_EYECATCHER + 2,	dump_application_data,	search_application,	{ 1, 0, 0 },	0 }
	
	,{ DICT_TYPE,		"TYPE",		sizeof(dict_type_data_t),	1, 	DICT_APPLICATION,
			OBJECT_EYECATCHER + 3,	dump_type_data,		search_type,		{ 1, 0, 0 },	0 }
	
	,{ DICT_TYPE_ENUM,	"TYPE_ENUM",	sizeof(dict_type_enum_data_t),	2, 	DICT_TYPE,
			OBJECT_EYECATCHER + 4,	NULL,			search_type_enum,	{ 1, 1, 0 },	0 }
	
	,{ DICT_AVP,		"AVP",		sizeof(dict_avp_data_t),	1, 	DICT_TYPE,
			OBJECT_EYECATCHER + 5,	dump_avp_data,		search_avp,		{ 1, 1,	0 },	0 }
	
	,{ DICT_COMMAND,	"COMMAND",	sizeof(dict_cmd_data_t),	1, 	DICT_APPLICATION,
			OBJECT_EYECATCHER + 6,	dump_command_data,	search_cmd,		{ 1, 1, 0 },	0 }
	
	,{ DICT_RULE,		"RULE",		sizeof(dict_rule_data_t),	2, 	-1 /* special case: grouped avp or command */,
			OBJECT_EYECATCHER + 7,	dump_rule_data,		search_rule,		{ 1, 0, 0 },	0 }
	
};
	
/* Macro to verify a "type" value */
#define CHECK_TYPE( type ) ( ((type) > 0) && ((type) <= DICT_TYPE_MAX) )

/* Get information line for a given object */
#define _OBINFO(object) (dict_obj_info[CHECK_TYPE(_O(object)->type) ? _O(object)->type : 0])

/*******************************************************************************************************/
/*******************************************************************************************************/
/*                                                                                                     */
/*                                  Lists management functions                                         */
/*                                                                                                     */
/*******************************************************************************************************/
/*******************************************************************************************************/

/* These functions are useful to manage the lists elements of the objects. 
 * The functions used to order the lists are defined later in this file (order_*).
 */

/* Initialize a list */
static void init_list ( _dict_list_t * list, _dict_object_t * o )
{
	list->next = list;
	list->prev = list;
	list->o = o;
}

/* Insert an object in a list. The write-lock must already have been acquired */
/* 0: OK, EEXIST: element already exists. */
static int insert_in_list ( 	_dict_list_t * obj_l, 
				_dict_list_t * sentinel, 
				int (*cmp_fct)(_dict_object_t *, _dict_object_t *), 
				_dict_object_t ** ref_duplicate )
{
	_dict_list_t * li = sentinel;
	int cmp;
	
	/* Some debug sanity checks */
	assert(obj_l->o != NULL);
	assert(obj_l->next == obj_l);
	assert(obj_l->prev == obj_l);
	
	/* loop in the list */
	while (li->next != sentinel)
	{
		/* Compare the object to insert with the next object in list */
		cmp = cmp_fct( obj_l->o, li->next->o );
		if (!cmp) {
			/* An element with the same key already exists */
			if (ref_duplicate != NULL)
				*ref_duplicate = li->next->o;
			return EEXIST;
		}
		
		if (cmp < 0)
			break; /* We must insert the element here */
		
		li = li->next;
	}
	
	/* Now insert the element between li and li->next */
	obj_l->next = li->next;
	obj_l->prev = li;
	li->next->prev = obj_l;
	li->next = obj_l;
	
	/* Ok */
	return 0;
}

/* Remove an object from a list. The write-lock must already have been acquired */
static void remove_from_list ( _dict_list_t * obj_l )
{
	/* Some debug sanity checks */
	assert(obj_l->o != NULL);
	assert(obj_l->next != obj_l);
	assert(obj_l->prev != obj_l);
	
	/* Now remove the element */
	obj_l->next->prev = obj_l->prev;
	obj_l->prev->next = obj_l->next;
	obj_l->prev = obj_l;
	obj_l->next = obj_l;
}

/* Forward declaration */
static void destroy_object(_dict_object_t * obj);

/* Destroy all objects in a list - the lock must be held */
static void destroy_list(_dict_list_t * sentinel) 
{
	_dict_list_t * li;
	
	/* Some debug sanity checks */
	assert(sentinel->o == NULL);
	
	/* loop in the list */
	while ((li = sentinel->next) != sentinel)
	{
		/* When destroying the object, it is unlinked from the list */
		destroy_object(li->o);
	}
}
	


/*******************************************************************************************************/
/*******************************************************************************************************/
/*                                                                                                     */
/*                                  Objects management                                                 */
/*                                                                                                     */
/*******************************************************************************************************/
/*******************************************************************************************************/

/* Functions to manage the objects creation and destruction. */


/* Initialize an object */
static void init_object( _dict_object_t * obj, dict_object_type_t type )
{
	int i;
	/* Clean the object first */
	memset ( obj, 0, sizeof(_dict_object_t));
	
	if (!CHECK_TYPE(type))
		return;

	obj->type = type;
	obj->objeyec = OBJECT_EYECATCHER;
	obj->typeyec = _OBINFO(obj).eyecatcher;

	/* We don't initialize the data nor the parent here */
	
	/* Now init the lists */
	for (i=0; i<NB_LISTS_PER_OBJ; i++) {
		if (_OBINFO(obj).haslist[i] != 0) 
			init_list(&obj->list[i], obj);
		else
			init_list(&obj->list[i], NULL);
	}
}

#define DUP_string( str ) {				\
	char * __str = (str);				\
	str = NULL;					\
	if (!(str = strdup(__str))) {			\
		TRACE_DEBUG(INFO, 			\
			"strdup failed: %s. ENOMEM", 	\
			strerror(errno));		\
		return ENOMEM;				\
	}						\
}		

/* Initialize the "data" part of an object */
int init_object_data(void * dest, void * source, dict_object_type_t type)
{
	int ret = 0;

	/* Generic: copy the full data structure */	
	memcpy( dest, source, dict_obj_info[type].datasize );
	
	/* Then strings must be duplicated, not copied */
	/* This function might be simplified by always defining the "name" field as the first field of the structures, but... it's error-prone */
	switch (type) {
		case DICT_VENDOR:
			DUP_string( ((dict_vendor_data_t *)dest)->vendor_name );
			break;
		
		case DICT_APPLICATION:
			DUP_string( ((dict_application_data_t *)dest)->application_name );
			break;
			
		case DICT_TYPE:
			DUP_string( ((dict_type_data_t *)dest)->type_name );
			break;
			
		case DICT_TYPE_ENUM:
			DUP_string( ((dict_type_enum_data_t *)dest)->enum_name );
			/* Do we have to copy the value in case of octetstring constants? ... */
			break;

		case DICT_AVP:
			DUP_string( ((dict_avp_data_t *)dest)->avp_name );
			break;
			
		case DICT_COMMAND:
			DUP_string( ((dict_cmd_data_t *)dest)->cmd_name );
			break;
	}
	
	return 0;
}


/* Check that an object is valid (0: OK, EINVAL: error) */
static int verify_object( _dict_object_t * obj )
{
	int ret = 0;
	
	if (  (!obj) 
	   || (obj->objeyec != OBJECT_EYECATCHER)
	   || (!CHECK_TYPE(obj->type))
	   || (obj->typeyec != dict_obj_info[obj->type].eyecatcher)
	   ) {
		if (obj) {
			TRACE_DEBUG(FULL, "obj->objeyec : %x - %x\n", obj->objeyec, OBJECT_EYECATCHER);
			TRACE_DEBUG(FULL, "obj->type    : %d\n", obj->type);
			TRACE_DEBUG(FULL, "obj->objeyec : %x - %x\n", obj->objeyec, OBJECT_EYECATCHER);
			TRACE_DEBUG(FULL, "obj->typeyec : %x - %x\n", obj->typeyec, _OBINFO(obj).eyecatcher);
		}
		return EINVAL;
	}
	
	/* The object is probably valid. */
	return 0;
}

/* Free the data associated to an object */
static void destroy_object_data(_dict_object_t * obj)
{
	switch (obj->type) {
		case DICT_VENDOR:
			free( obj->data.vendor.vendor_name );
			break;
		
		case DICT_APPLICATION:
			free( obj->data.application.application_name );
			break;
			
		case DICT_TYPE:
			free( obj->data.type.type_name );
			break;
			
		case DICT_TYPE_ENUM:
			free( obj->data.type_enum.enum_name );
			/* Do we have to free the value in case of octetstring constants? ... */
			break;

		case DICT_AVP:
			free( obj->data.avp.avp_name );
			break;
			
		case DICT_COMMAND:
			free( obj->data.cmd.cmd_name );
			break;
	}
}

/* Free an object and its sublists */
static void destroy_object(_dict_object_t * obj)
{
	int i;
	
	/* Update global count */
	_OBINFO(obj).count--;
	
	/* Mark the object as invalid */
	obj->objeyec = 0xdead;
	
	/* First, destroy the data associated to the object */
	destroy_object_data(obj);
	
	for (i=0; i<NB_LISTS_PER_OBJ; i++) {
		if (_OBINFO(obj).haslist[i])
			/* unlink the element from the list */
			remove_from_list( &obj->list[i] );
		else
			/* This is either a sentinel or unused (=emtpy) list, let's destroy it */
			destroy_list( &obj->list[i] );
	}
	
	/* Last, destroy the object */
	free(obj);
}


/*******************************************************************************************************/
/*******************************************************************************************************/
/*                                                                                                     */
/*                                  Compare functions                                                  */
/*                                                                                                     */
/*******************************************************************************************************/
/*******************************************************************************************************/

/* Functions that are used to ordering the lists. */


/* Compare two values */
#define order_scalar( i1, i2 ) \
	((i1 < i2 ) ? -1 : ( i1 > i2 ? 1 : 0 )) 

/* Compare two vendor objects by their id (checks already performed) */
static int order_vendor_by_id ( _dict_object_t *o1, _dict_object_t *o2 )
{
	return order_scalar( o1->data.vendor.vendor_id, o2->data.vendor.vendor_id );
}

/* Compare two application objects by their id (checks already performed) */
static int order_appli_by_id  ( _dict_object_t *o1, _dict_object_t *o2 )
{
	return order_scalar( o1->data.application.application_id, o2->data.application.application_id );
}

/* Compare two type objects by their name (checks already performed) */
static int order_type_by_name ( _dict_object_t *o1, _dict_object_t *o2 )
{
	return strcmp( o1->data.type.type_name, o2->data.type.type_name );
}

/* Compare two type_enum objects by their names (checks already performed) */
static int order_enum_by_name ( _dict_object_t *o1, _dict_object_t *o2 )
{
	return strcmp( o1->data.type_enum.enum_name, o2->data.type_enum.enum_name );
}

/* Compare two type_enum objects by their values (checks already performed) */
static int order_enum_by_val  ( _dict_object_t *o1, _dict_object_t *o2 )
{
	size_t oslen;
	int cmp = 0;
	
	/* The comparison function depends on the type of data */
	switch ( o1->parent->data.type.type_base ) {
		case AVP_TYPE_OCTETSTRING:
			oslen = o1->data.type_enum.enum_value.os.len;
			if (o2->data.type_enum.enum_value.os.len < oslen)
				oslen = o2->data.type_enum.enum_value.os.len;
			cmp = memcmp(o1->data.type_enum.enum_value.os.data, o2->data.type_enum.enum_value.os.data, oslen );
			return (cmp ? cmp : order_scalar(o1->data.type_enum.enum_value.os.len,o2->data.type_enum.enum_value.os.len));
		
		case AVP_TYPE_INTEGER32:
			return order_scalar( o1->data.type_enum.enum_value.i32, o2->data.type_enum.enum_value.i32 );

		case AVP_TYPE_INTEGER64:
			return order_scalar( o1->data.type_enum.enum_value.i64, o2->data.type_enum.enum_value.i64 );

		case AVP_TYPE_UNSIGNED32:
			return order_scalar( o1->data.type_enum.enum_value.u32, o2->data.type_enum.enum_value.u32 );

		case AVP_TYPE_UNSIGNED64:
			return order_scalar( o1->data.type_enum.enum_value.u64, o2->data.type_enum.enum_value.u64 );

		case AVP_TYPE_FLOAT32:
			return order_scalar( o1->data.type_enum.enum_value.f32, o2->data.type_enum.enum_value.f32 );

		case AVP_TYPE_FLOAT64:
			return order_scalar( o1->data.type_enum.enum_value.f64, o2->data.type_enum.enum_value.f64 );

		case AVP_TYPE_GROUPED:
		default:
			assert(0);
	}
	return 0;
}

/* Compare two avp objects by their codes (checks already performed) */
static int order_avp_by_code  ( _dict_object_t *o1, _dict_object_t *o2 )
{
	return order_scalar( o1->data.avp.avp_code, o2->data.avp.avp_code );
}

/* Compare two avp objects by their names (checks already performed) */
static int order_avp_by_name  ( _dict_object_t *o1, _dict_object_t *o2 )
{
	return strcmp( o1->data.avp.avp_name, o2->data.avp.avp_name );
}

/* Compare two command objects by their names (checks already performed) */
static int order_cmd_by_name  ( _dict_object_t *o1, _dict_object_t *o2 )
{
	return strcmp( o1->data.cmd.cmd_name, o2->data.cmd.cmd_name );
}

/* Compare two command objects by their codes and flags (request or answer) (checks already performed) */
static int order_cmd_by_codefl( _dict_object_t *o1, _dict_object_t *o2 )
{
	uint8_t fl1, fl2;
	int ret = order_scalar( o1->data.cmd.cmd_code, o2->data.cmd.cmd_code );
	if (ret) 
		return ret;
	/* Same code, must compare the value of the 'R' flag */
	fl1 = o1->data.cmd.cmd_flag_val & CMD_FLAG_REQUEST;
	fl2 = o2->data.cmd.cmd_flag_val & CMD_FLAG_REQUEST;
	
	/* We want requests first, so we reverse the operators here */
	return order_scalar(fl2, fl1);
		
}

/* Compare two rule object by the AVP name that they refer (checks already performed) */
static int order_rule_by_avpn ( _dict_object_t *o1, _dict_object_t *o2 )
{
	return strcmp( _O(o1->data.rule.rule_avp)->data.avp.avp_name, _O(o2->data.rule.rule_avp)->data.avp.avp_name );
}

/*******************************************************************************************************/
/*******************************************************************************************************/
/*                                                                                                     */
/*                                  Search  functions                                                  */
/*                                                                                                     */
/*******************************************************************************************************/
/*******************************************************************************************************/

/* Functions used to search for objects in the lists, according to some criteria */

/* The following macros assume that "what", "ret", "result" (variables), and "end" (label) exist
in the local context where they are called. They are meant to be called only inside the functions that follow. */

/* For searchs of type "xxx_OF_xxx": children's parent or default parent */
#define SEARCH_childs_parent( type_of_child, default_parent ) {			\
	_dict_object_t *__child = (_dict_object_t *) what;			\
	if ( (verify_object(__child) != 0) 					\
	     || (__child->type != (type_of_child))) {				\
		TRACE_DEBUG( INFO, "Invalid object passed" );			\
		ret = EINVAL;							\
		goto end;							\
	}									\
	ret = 0;								\
	if (result)								\
		*result = (__child->parent ? __child->parent :(default_parent));\
}

/* For search of strings in lists. isindex= 1 if the string is the ordering key of the list */
#define SEARCH_string( str, sentinel, datafield, isindex ) {			\
	char * __str = (char *) str;						\
	int __cmp;								\
	_dict_list_t * __li;							\
	for  (__li = (sentinel); __li->next != (sentinel); __li = __li->next) {	\
		__cmp = strcmp(__str, __li->next->o->data. datafield );		\
		if (__cmp == 0) {						\
			if (result)						\
				*result = __li->next->o;			\
			ret = 0;						\
			goto end;						\
		}								\
		if ((isindex) && (__cmp < 0))					\
			break;							\
	}									\
	ret = ENOENT;								\
}
/* For search of octetstrings in lists (not \0 terminated). */
#define SEARCH_ocstring( ostr, length, sentinel, osdatafield, isindex ) {	\
	unsigned char * __ostr = (unsigned char *) ostr;			\
	int __cmp;								\
	size_t __len;								\
	_dict_list_t * __li;							\
	for  (__li = (sentinel); __li->next != (sentinel); __li = __li->next) {	\
		__len = __li->next->o->data. osdatafield .len;			\
		if ( __len > (length) )						\
			__len = (length);					\
		__cmp = memcmp(__ostr, 						\
				__li->next->o->data. osdatafield .data, 	\
				__len);						\
		if (! __cmp) {							\
			__cmp = order_scalar( length,				\
					__li->next->o->data. osdatafield .len); \
		}								\
		if (__cmp == 0) {						\
			if (result)						\
				*result = __li->next->o;			\
			ret = 0;						\
			goto end;						\
		}								\
		if ((isindex) && (__cmp < 0))					\
			break;							\
	}									\
	ret = ENOENT;								\
}
/* For search of AVP name in rule lists. */
#define SEARCH_ruleavpname( str, sentinel ) {					\
	char * __str = (char *) str;						\
	int __cmp;								\
	_dict_list_t * __li;							\
	for  (__li = (sentinel); __li->next != (sentinel); __li = __li->next) {	\
		__cmp = strcmp(__str, 						\
		   _O(__li->next->o->data.rule.rule_avp)->data.avp.avp_name);	\
		if (__cmp == 0) {						\
			if (result)						\
				*result = __li->next->o;			\
			ret = 0;						\
			goto end;						\
		}								\
		if (__cmp < 0)							\
			break;							\
	}									\
	ret = ENOENT;								\
}

/* For search of scalars in lists. isindex= 1 if the value is the ordering key of the list */
#define SEARCH_scalar( value, sentinel, datafield, isindex, defaultobj ) {	\
	int __cmp;								\
	_dict_list_t * __li;							\
	if ((defaultobj) && ((defaultobj)->data. datafield  == value)) {	\
		if (result)							\
			*result = (defaultobj);					\
		ret = 0;							\
		goto end;							\
	}									\
	for  (__li = (sentinel); __li->next != (sentinel); __li = __li->next) {	\
		__cmp = order_scalar(value, __li->next->o->data. datafield );	\
		if (__cmp == 0) {						\
			if (result)						\
				*result = __li->next->o;			\
			ret = 0;						\
			goto end;						\
		}								\
		if ((isindex) && (__cmp < 0))					\
			break;							\
	}									\
	ret = ENOENT;								\
}

/* For search of commands in lists by code and flag. R_flag_val = 0 or CMD_FLAG_REQUEST */
#define SEARCH_codefl( value, R_flag_val) {					\
	int __cmp;								\
	_dict_list_t * __li;							\
	for  (	  __li = &g_list_cmd_code; 					\
		  __li->next != &g_list_cmd_code; 				\
		  __li = __li->next) {						\
		__cmp = order_scalar(value, __li->next->o->data.cmd.cmd_code );	\
		if (__cmp == 0) {						\
			uint8_t __mask, __val;					\
			__mask = __li->next->o->data.cmd.cmd_flag_mask;		\
			__val  = __li->next->o->data.cmd.cmd_flag_val;		\
			if ( ! (__mask & CMD_FLAG_REQUEST))			\
				continue;					\
			if ( ( __val & CMD_FLAG_REQUEST ) != R_flag_val )	\
				continue;					\
			if (result)						\
				*result = __li->next->o;			\
			ret = 0;						\
			goto end;						\
		}								\
		if (__cmp < 0)							\
			break;							\
	}									\
	ret = ENOENT;								\
}


static int search_vendor ( int criteria, void * what, _dict_object_t **result )
{
	int ret = 0;
	vendor_id_t id;
	
	switch (criteria) {
		case VENDOR_BY_ID_REF:
		case VENDOR_BY_ID_VAL:
			if (criteria == VENDOR_BY_ID_REF) {
				id = *(vendor_id_t *) what;
			} else {
				assert(sizeof(vendor_id_t) <= sizeof(void *));  /* otherwise the cast does not work */
				id = (vendor_id_t) what;
			}

			SEARCH_scalar( id, &g_dict_vendors.list[0], vendor.vendor_id, 1, &g_dict_vendors );
			break;
				
		case VENDOR_BY_NAME:
			/* "what" is a vendor name */
			SEARCH_string( what, &g_dict_vendors.list[0], vendor.vendor_name, 0);
			break;
			
		case VENDOR_OF_APPLICATION:
			/* "what" should be an application object */
			SEARCH_childs_parent( DICT_APPLICATION, &g_dict_vendors );
			break;
		
		default:
			/* Invalid criteria */
			TRACE_DEBUG( INFO, "Invalid criteria: %d. EINVAL.", criteria );
			ret = EINVAL;
	}
end:
	return ret;
}

static int search_application ( int criteria, void * what, _dict_object_t **result )
{
	int ret = 0;
	application_id_t id;
	
	switch (criteria) {
		case APPLICATION_BY_ID_REF:
		case APPLICATION_BY_ID_VAL:
			if (criteria == APPLICATION_BY_ID_REF) {
				id = *(application_id_t *) what;
			} else {
				assert(sizeof(application_id_t) <= sizeof(void *));  /* otherwise the cast does not work */
				id = (application_id_t) what;
			}

			SEARCH_scalar( id, &g_dict_applications.list[0],  application.application_id, 1, &g_dict_applications );
			break;
				
		case APPLICATION_BY_NAME:
			/* "what" is an application name */
			SEARCH_string( what, &g_dict_applications.list[0], application.application_name, 0);
			break;
			
		case APPLICATION_OF_TYPE:
			/* "what" should be a type object */
			SEARCH_childs_parent( DICT_TYPE, &g_dict_applications );
			break;
		
		case APPLICATION_OF_COMMAND:
			/* "what" should be a command object */
			SEARCH_childs_parent( DICT_COMMAND, &g_dict_applications );
			break;
		
		default:
			/* Invalid criteria */
			TRACE_DEBUG( INFO, "Invalid criteria: %d. EINVAL.", criteria );
			ret = EINVAL;
	}
end:
	return ret;
}

static int search_type ( int criteria, void * what, _dict_object_t **result )
{
	int ret = 0;
	
	switch (criteria) {
		case TYPE_BY_NAME:
			/* "what" is a type name */
			SEARCH_string( what, &g_list_types, type.type_name, 1);
			break;
			
		case TYPE_OF_ENUM:
			/* "what" should be a type_enum object */
			SEARCH_childs_parent( DICT_TYPE_ENUM, NULL );
			break;
		
		case TYPE_OF_AVP:
			/* "what" should be an avp object */
			SEARCH_childs_parent( DICT_AVP, NULL );
			break;
		
				
		default:
			/* Invalid criteria */
			TRACE_DEBUG( INFO, "Invalid criteria: %d. EINVAL.", criteria );
			ret = EINVAL;
	}
end:
	return ret;
}

static int search_type_enum ( int criteria, void * what, _dict_object_t **result )
{
	int ret = 0;
	
	switch (criteria) {
		case ENUM_BY_STRUCT:
			{
				_dict_object_t * parent = NULL;
				dict_type_enum_request_t * _what = (dict_type_enum_request_t *) what;
				if (   (_what == NULL)
				    || ((_what->type_obj == NULL) && (_what->type_name == NULL)) )
				{
					TRACE_DEBUG( INFO, "Invalid dict_type_enum_request_t passed" );
					ret = EINVAL;
					goto end;
				}
				
				if (_what->type_obj != NULL) {
					parent = _O(_what->type_obj);
					if (verify_object(parent) || (parent->type != DICT_TYPE) ) {
						TRACE_DEBUG( INFO, "Invalid type_obj passed" );
						ret = EINVAL;
						goto end;
					}
				} else {
					/* We received only the type name, we must find it first */
					ret = search_type( TYPE_BY_NAME, _what->type_name, &parent );
					if (ret != 0) {
						TRACE_DEBUG( INFO, "Invalid type_name passed (%s)", strerror(ret) );
						ret = EINVAL;
						goto end;
					}
				}
				
				/* From here the "parent" object is valid */
				
				if ( _what->search.enum_name != NULL ) {
					/* We are looking for this string */
					SEARCH_string(  _what->search.enum_name, &parent->list[1], type_enum.enum_name, 1 );
				} else {
					/* We are looking for the value in enum_value */
					switch (parent->data.type.type_base) {
						case AVP_TYPE_OCTETSTRING:
							SEARCH_ocstring( _what->search.enum_value.os.data, 
									 _what->search.enum_value.os.len, 
									 &parent->list[2], 
									 type_enum.enum_value.os , 
									 1 );
							break;

						case AVP_TYPE_INTEGER32:
							SEARCH_scalar(	_what->search.enum_value.i32,
									&parent->list[2],
									type_enum.enum_value.i32,
									1,
									(_dict_object_t *)NULL);
							break;
							
						case AVP_TYPE_INTEGER64:
							SEARCH_scalar(	_what->search.enum_value.i64,
									&parent->list[2],
									type_enum.enum_value.i64,
									1,
									(_dict_object_t *)NULL);
							break;
							
						case AVP_TYPE_UNSIGNED32:
							SEARCH_scalar(	_what->search.enum_value.u32,
									&parent->list[2],
									type_enum.enum_value.u32,
									1,
									(_dict_object_t *)NULL);
							break;
							
						case AVP_TYPE_UNSIGNED64:
							SEARCH_scalar(	_what->search.enum_value.u64,
									&parent->list[2],
									type_enum.enum_value.u64,
									1,
									(_dict_object_t *)NULL);
							break;
							
						case AVP_TYPE_FLOAT32:
							SEARCH_scalar(	_what->search.enum_value.f32,
									&parent->list[2],
									type_enum.enum_value.f32,
									1,
									(_dict_object_t *)NULL);
							break;
							
						case AVP_TYPE_FLOAT64:
							SEARCH_scalar(	_what->search.enum_value.f64,
									&parent->list[2],
									type_enum.enum_value.f64,
									1,
									(_dict_object_t *)NULL);
							break;
							
						default:
							/* Invalid parent type basetype */
							TRACE_DEBUG( INFO, "Invalid parent type: (%s). EINVAL.", type_base_name[parent->data.type.type_base] );
							ret = EINVAL;
					}
				}
				
			}
			break;
		
				
		default:
			/* Invalid criteria */
			TRACE_DEBUG( INFO, "Invalid criteria: %d. EINVAL.", criteria );
			ret = EINVAL;
	}
end:
	return ret;
}

static int search_avp ( int criteria, void * what, _dict_object_t **result )
{
	int ret = 0;
	
	switch (criteria) {
		case AVP_BY_CODE_REF:
		case AVP_BY_CODE_VAL:
			{
				avp_code_t code;
				if (criteria == AVP_BY_CODE_REF) {
					code = *(avp_code_t *) what;
				} else {
					assert(sizeof(avp_code_t) <= sizeof(void *));  /* otherwise the cast does not work */
					code = (avp_code_t) what;
				}

				SEARCH_scalar( code, &g_dict_vendors.list[1],  avp.avp_code, 1, (_dict_object_t *)NULL );
			}
			break;
				
		case AVP_BY_NAME:
			/* "what" is the AVP name, vendor 0 */
			SEARCH_string( what, &g_dict_vendors.list[2], avp.avp_name, 1);
			break;
			
		case AVP_BY_CODE_AND_VENDOR:
		case AVP_BY_NAME_AND_VENDOR:
			{
				dict_avp_request_t * _what = (dict_avp_request_t *) what;
				_dict_object_t * vendor = NULL;
				
				if ((criteria == AVP_BY_NAME_AND_VENDOR) && (_what->avp_name == NULL)) {
					TRACE_DEBUG( INFO, "Invalid dict_avp_request_t passed, missing avp_name" );
					ret = EINVAL;
					goto end;
				}
				
				/* Now look for the vendor first */
				ret = search_vendor( VENDOR_BY_ID_REF, &_what->avp_vendor, &vendor );
				if (ret != 0) {
					TRACE_DEBUG( INFO, "Invalid dict_avp_request_t passed, vendor not found (%d -> %d '%s')", _what->avp_vendor, ret, strerror(ret) );
					ret = EINVAL;
					goto end;
				}
				
				/* We now have our vendor = head of the appropriate avp list */
				if (criteria == AVP_BY_NAME_AND_VENDOR) {
					SEARCH_string( _what->avp_name, &vendor->list[2], avp.avp_name, 1);
				} else {
					/* AVP_BY_CODE_AND_VENDOR */
					SEARCH_scalar( _what->avp_code, &vendor->list[1],  avp.avp_code, 1, (_dict_object_t *)NULL );
				}
			}
			break;
		
		default:
			/* Invalid criteria */
			TRACE_DEBUG( INFO, "Invalid criteria: %d. EINVAL.", criteria );
			ret = EINVAL;
	}
end:
	return ret;
}

static int search_cmd ( int criteria, void * what, _dict_object_t **result )
{
	int ret = 0;
	
	switch (criteria) {
		case CMD_BY_NAME:
			/* "what" is a command name */
			SEARCH_string( what, &g_list_cmd_name, cmd.cmd_name, 1);
			break;
			
		case CMD_BY_CODE_R_VAL:
		case CMD_BY_CODE_A_VAL:
		case CMD_BY_CODE_R_REF:
		case CMD_BY_CODE_A_REF:
			{
				command_code_t code;
				uint8_t searchfl = 0;
				
				/* The command code that we are searching */
				if ( (criteria == CMD_BY_CODE_R_REF) || (criteria == CMD_BY_CODE_A_REF)) {
					code = *(command_code_t *) what;
				} else {
					assert(sizeof(command_code_t) <= sizeof(void *));  /* otherwise the cast does not work */
					code = (command_code_t) what;
				}
				
				/* The flag (request or answer) of the command we are searching */
				if ( (criteria == CMD_BY_CODE_R_VAL) || (criteria == CMD_BY_CODE_R_REF)) {
					searchfl = CMD_FLAG_REQUEST;
				}
				
				/* perform the search */
				SEARCH_codefl( code, searchfl );
			}
			break;
				
		case CMD_ANSWER:
			{
				/* "what" is a command object of type "request" */
				_dict_object_t * req = (_dict_object_t *) what;
				_dict_object_t * ans = NULL;
				
				if (  verify_object(req) 
				   || (req->type != DICT_COMMAND)
				   || (!(req->data.cmd.cmd_flag_mask & CMD_FLAG_REQUEST))
				   || (!(req->data.cmd.cmd_flag_val  & CMD_FLAG_REQUEST)) ) {
					TRACE_DEBUG( INFO, "Invalid command object passed" );
					ret = EINVAL;
					goto end;
				}
				
				/* The answer is supposed to be the next element in the list, if it exists */
				ans = req->list[1].next->o;
				if ( ans == NULL ) {
					TRACE_DEBUG( FULL, "the request was the last element in the list" );
					ret = ENOENT;
					goto end;
				}
				
				/* Now check that the ans element is really the correct one */
				if (  (ans->data.cmd.cmd_code != req->data.cmd.cmd_code)
				   || (!(ans->data.cmd.cmd_flag_mask & CMD_FLAG_REQUEST))
				   || (  ans->data.cmd.cmd_flag_val  & CMD_FLAG_REQUEST ) ) {
					TRACE_DEBUG( FULL, "the answer does not follow the request in the list" );
					ret = ENOENT;
					goto end;
				}
				
				if (result)
					*result = ans;
				ret = 0;
			}						
			break;
			
		default:
			/* Invalid criteria */
			TRACE_DEBUG( INFO, "Invalid criteria: %d. EINVAL.", criteria );
			ret = EINVAL;
	}
end:
	return ret;
}

static int search_rule ( int criteria, void * what, _dict_object_t **result )
{
	int ret = 0;
	
	switch (criteria) {
		case RULE_BY_AVP_AND_PARENT:
			{
				_dict_object_t * parent = NULL;
				_dict_object_t * avp = NULL;
				dict_rule_request_t * _what = (dict_rule_request_t *) what;
				
				if (   (_what == NULL)
				    || ((parent = _O(_what->rule_parent)) == NULL)
				    || ((avp    = _O(_what->rule_avp   )) == NULL))
				{
					TRACE_DEBUG( INFO, "Invalid dict_rule_request_t passed" );
					ret = EINVAL;
					goto end;
				}
				
				if (	  verify_object(parent) 
					|| (      (parent->type != DICT_COMMAND)
					    && (  (parent->type != DICT_AVP)  
						||(parent->data.avp.avp_basetype != AVP_TYPE_GROUPED) ) ) ) {
					TRACE_DEBUG( INFO, "Invalid rule_parent in dict_rule_request_t" );
					ret = EINVAL;
					goto end;
				}
				
				if ( verify_object(avp) || (avp->type != DICT_AVP)  ) {
					TRACE_DEBUG( INFO, "Invalid rule_avp in dict_rule_request_t" );
					ret = EINVAL;
					goto end;
				}
				
				/* Perform the search */
				SEARCH_ruleavpname( avp->data.avp.avp_name, &parent->list[2]);
				
			}
			break;
			
		default:
			/* Invalid criteria */
			TRACE_DEBUG( INFO, "Invalid criteria: %d. EINVAL.", criteria );
			ret = EINVAL;
	}
end:
	return ret;
}


/*******************************************************************************************************/
/*******************************************************************************************************/
/*                                                                                                     */
/*                                  Dump / debug functions                                             */
/*                                                                                                     */
/*******************************************************************************************************/
/*******************************************************************************************************/

/* The following functions are used to debug the module, and allow to print out the content of the dictionary */
static void dump_vendor_data ( void * data )
{
	dict_vendor_data_t * vendor = (dict_vendor_data_t *)data;
	
	log_debug("data: %-6u \"%s\"", vendor->vendor_id, vendor->vendor_name);
}
static void dump_application_data ( void * data )
{
	dict_application_data_t * appli = (dict_application_data_t *) data;
	log_debug("data: %-6u \"%s\"", appli->application_id, appli->application_name);
}
static void dump_type_data ( void * data )
{
	dict_type_data_t * type = ( dict_type_data_t * ) data;
	
	log_debug("data: %-12s \"%s\"", 
			type_base_name[type->type_base], 
			type->type_name);
}
static void dump_type_enum_data ( dict_type_enum_data_t * type_enum, dict_base_type_t type )
{
	const int LEN_MAX = 20;
	log_debug("data: (%-12s) \"%s\" -> ", type_base_name[type], type_enum->enum_name);
	switch (type) {
		case AVP_TYPE_OCTETSTRING:
			{
				int i, n=LEN_MAX;
				if (type_enum->enum_value.os.len < LEN_MAX)
					n = type_enum->enum_value.os.len;
				for (i=0; i < n; i++)
					log_debug("0x%02.2X/'%c' ", type_enum->enum_value.os.data[i], type_enum->enum_value.os.data[i]);
				if (n == LEN_MAX)
					log_debug("...");
			}
			break;
		
		case AVP_TYPE_INTEGER32:
			log_debug("%i", type_enum->enum_value.i32);
			break;

		case AVP_TYPE_INTEGER64:
			log_debug("%lli", type_enum->enum_value.i64);
			break;

		case AVP_TYPE_UNSIGNED32:
			log_debug("%u", type_enum->enum_value.u32);
			break;

		case AVP_TYPE_UNSIGNED64:
			log_debug("%llu", type_enum->enum_value.u64);
			break;

		case AVP_TYPE_FLOAT32:
			log_debug("%g", (double)type_enum->enum_value.f32);
			break;

		case AVP_TYPE_FLOAT64:
			log_debug("%Lg", type_enum->enum_value.f64);
			break;
		
		default:
			log_debug("??? (ERROR unknown type %d)", type);
	}
}
static void dump_avp_data ( void * data )
{
	dict_avp_data_t * avp = (dict_avp_data_t * ) data;
	log_debug("data: v/m:" DUMP_AVPFL_str "/" DUMP_AVPFL_str ", %12s, %-6u \"%s\"", 
			DUMP_AVPFL_val(avp->avp_flag_val), 
			DUMP_AVPFL_val(avp->avp_flag_mask), 
			type_base_name[avp->avp_basetype], 
			avp->avp_code, 
			avp->avp_name );
}
static void dump_command_data ( void * data )
{
	dict_cmd_data_t * cmd = (dict_cmd_data_t *) data;
	log_debug("data: v/m:" DUMP_CMDFL_str "/" DUMP_CMDFL_str ", %-6u \"%s\"", 
			DUMP_CMDFL_val(cmd->cmd_flag_val), DUMP_CMDFL_val(cmd->cmd_flag_mask), cmd->cmd_code, cmd->cmd_name);
}
static void dump_rule_data ( void * data )
{
	dict_rule_data_t * rule = (dict_rule_data_t * )data;
	log_debug("data: pos:%d ord:%d m/M/t:%2d/%2d/%2d avp:\"%s\"", 
			rule->rule_position, 
			rule->rule_order, 
			rule->rule_min, 
			rule->rule_max,
			rule->rule_template, 
			_O(rule->rule_avp)->data.avp.avp_name);
}

static void dump_object ( _dict_object_t * obj, int parents, int depth, int indent );

static void dump_list ( _dict_list_t * sentinel, int parents, int depth, int indent )
{
	_dict_list_t * li = sentinel;
	/* We don't lock here, the caller must have taken the dictionary lock for reading already */
	while (li->next != sentinel)
	{
		li = li->next;
		dump_object( li->o, parents, depth, indent );
	}
}

static void dump_object ( _dict_object_t * obj, int parents, int depth, int indent )
{
	if (obj == NULL)
		return;
	
	if (parents)
		dump_object (obj->parent, parents-1, 0, indent + 1 );
	
	log_debug("%*s@%p: %s%s (p:%-9p) ", 
			indent,
			"",
			obj, 
			verify_object(obj) ? "NON VALID " : "", 
			_OBINFO(obj).name, 
			obj->parent);
	
	if (obj->type == DICT_TYPE_ENUM)
		dump_type_enum_data ( &obj->data.type_enum, _O(obj->parent)->data.type.type_base );
	else
		_OBINFO(obj).dump_data(&obj->data);
	
	log_debug("\n");
	
	if (depth) {
		int i;
		for (i=0; i<NB_LISTS_PER_OBJ; i++) {
			if ((obj->list[i].o == NULL) && (obj->list[i].next != &obj->list[i])) {
				log_debug("%*s>%p: list[%d]:\n", indent, "", obj, i);
				dump_list(&obj->list[i], parents, depth - 1, indent + 2);
			}
		}
	}
}

void dump_dictionary(void)
{
	int ret = 0, i;
	ret = pthread_rwlock_rdlock( &dict_lock );
	if (ret != 0) {
		TRACE_DEBUG (FULL, "pthread_rwlock_rdlock: %d (%s)", ret, strerror(ret) );
	}
	
	log_debug("######################################################\n");
	log_debug("###### Dumping vendors, AVPs and related rules #######\n");
	
	dump_object( &g_dict_vendors, 0, 3, 0 );
	
	log_debug("###### Dumping applications #######\n");

	dump_object( &g_dict_applications, 0, 1, 0 );
	
	log_debug("###### Dumping types #######\n");

	dump_list( &g_list_types, 0, 2, 0 );
	
	log_debug("###### Dumping commands per name #######\n");

	dump_list( &g_list_cmd_name, 0, 2, 0 );
	
	log_debug("###### Dumping commands per code and flags #######\n");

	dump_list( &g_list_cmd_code, 0, 0, 0 );
	
	log_debug("###### Statistics #######\n");

	for (i=1; i<=DICT_TYPE_MAX; i++)
		log_debug(" %5d objects of type %s\n", dict_obj_info[i].count, dict_obj_info[i].name);
	
	log_debug("######################################################\n");
	
			
	
	/* Free the rwlock */
	ret = pthread_rwlock_unlock( &dict_lock );
	if (ret != 0) {
		TRACE_DEBUG (FULL, "pthread_rwlock_unlock: %d (%s)", ret, strerror(ret) );
	}
}

void dict_dump( dict_object_t * object, int parent, int depth)
{
	dump_object ( _O(object), parent, depth, 1 );
}


/*******************************************************************************************************/
/*******************************************************************************************************/
/*                                                                                                     */
/*                                  The API functions                                                  */
/*                                                                                                     */
/*******************************************************************************************************/
/*******************************************************************************************************/

/* These are the functions exported to the extensions, and the main entry points of the module. */

/* Get the type of an object */
int dict_gettype ( dict_object_t * object, dict_object_type_t * type)
{
	if (!type || verify_object(_O(object))) {
		TRACE_DEBUG (FULL, "EINVAL." );
		return EINVAL;
	}
	
	/* Copy the type and return */
	*type = _O(object)->type;
	return 0;
}

/* Get the data associated to an object */
int dict_getval ( dict_object_t * object, void * val)
{
	if (!val || verify_object(_O(object))) {
		TRACE_DEBUG (FULL, "EINVAL." );
		return EINVAL;
	}
	
	/* Copy the value and return */
	memcpy(val, &_O(object)->data, _OBINFO(object).datasize);;
	return 0;
}

/* Add a new object in the dictionary */
int dict_new ( dict_object_type_t type, void * data, dict_object_t * parent, dict_object_t **ref )
{
	int ret = 0;
	_dict_object_t * new = NULL;
	_dict_object_t * vendor = NULL;
	
	TRACE_DEBUG (FULL, "dict_new (%s, %p, %p, %p).", dict_obj_info[CHECK_TYPE(type) ? type : 0].name, data, parent, ref );
	
	/* Check parameters */
	if (   !CHECK_TYPE(type) 
	    || !data) {
		TRACE_DEBUG (FULL, "EINVAL." );
		return EINVAL;
	}
	
	/* Check the "parent" parameter */
	switch (dict_obj_info[type].parent) {
		case 0:	/* parent is forbidden */
			if (parent != NULL) {
				TRACE_DEBUG (FULL, "Parent is forbidden, EINVAL." );
				return EINVAL;
			}
		
		case 1:	/* parent is optional */
			if (parent == NULL)
				break;
		
		case 2: /* parent is mandatory */
			if (verify_object(parent)) {
				TRACE_DEBUG (FULL, "Parent is invalid, EINVAL." );
				return EINVAL;
			}
			
			if (type == DICT_RULE ) { /* Special case : grouped AVP or Command parents are allowed */
				if (       (_O(parent)->type != DICT_COMMAND ) 
				    && (   (_O(parent)->type != DICT_AVP)
					|| (_O(parent)->data.avp.avp_basetype != AVP_TYPE_GROUPED ) ) )
					ret = EINVAL;
			} else {
				if ( _O(parent)->type != dict_obj_info[type].parenttype )
					ret = EINVAL;
			}
			if (ret != 0) {
				TRACE_DEBUG (FULL, "Parent is wrong type, EINVAL." );
				return EINVAL;
			}
	}
	
	/* For AVP object, we must also check that the "vendor" referenced exists */
	if (type == DICT_AVP) {
		ret = dict_search( DICT_VENDOR, VENDOR_BY_ID_REF, &(((dict_avp_data_t *)data)->avp_vendor), (dict_object_t **)&vendor );
		if (ret != 0) {
			TRACE_DEBUG (INFO, "Referenced vendor_id not found in dictionary, EINVAL." );
			return EINVAL;
		}
		
		/* Also check if a parent is provided, that the type are the same */
		if (parent && (_O(parent)->data.type.type_base != ((dict_avp_data_t *)data)->avp_basetype)) {
			TRACE_DEBUG (INFO, "Referenced avp_basetype does not match the parent's type, EINVAL." );
			return EINVAL;
		}
	}
	
	/* For COMMAND object, check that the 'R' flag is fixed */
	if ((type == DICT_COMMAND) && (! (((dict_cmd_data_t *)data)->cmd_flag_mask & CMD_FLAG_REQUEST))) {
		TRACE_DEBUG (INFO, "The CMD_FLAG_REQUEST must be fixed (in cmd_flag_mask), EINVAL." );
		return EINVAL;
	}
	
	/* Parameters are valid, create the new object */
	new = malloc(sizeof(_dict_object_t));
	if (!new) {
		log_error("Memory allocation failed. Memory full?\n");
		TRACE_DEBUG (INFO, "malloc failed." );
		return ENOMEM;
	}
	
	/* Initialize the data of the new object */
	init_object(new, type);
	init_object_data(&new->data, data, type);
	new->parent = parent;
	
	/* We will change the dictionary => acquire the write lock */
	ret = pthread_rwlock_wrlock(&dict_lock);
	if (ret != 0) {
		log_error("Unable to acquire a lock: %s\n", strerror(ret));
		TRACE_DEBUG (INFO, "pthread_rwlock_wrlock failed." );
		goto error_free;
	}
	
	/* Now link the object -- this also checks that no object with same keys already exists */
	switch (type) {
		case DICT_VENDOR:
			/* A vendor object is linked in the g_dict_vendors.list[0], by their id */
			ret = insert_in_list ( &new->list[0], &g_dict_vendors.list[0], order_vendor_by_id, (_dict_object_t **)ref );
			if (ret != 0) {
				TRACE_DEBUG(INFO, "Duplicate found. EEXIST");
				goto error_unlock;
			}
			break;
		
		case DICT_APPLICATION:
			/* An application object is linked in the g_dict_applciations.list[0], by their id */
			ret = insert_in_list ( &new->list[0], &g_dict_applications.list[0], order_appli_by_id, (_dict_object_t **)ref );
			if (ret != 0) {
				TRACE_DEBUG(INFO, "Duplicate found. EEXIST");
				goto error_unlock;
			}
			break;
		
		case DICT_TYPE:
			/* A type object is linked in g_list_types by its name */
			ret = insert_in_list ( &new->list[0], &g_list_types, order_type_by_name, (_dict_object_t **)ref );
			if (ret != 0) {
				TRACE_DEBUG(INFO, "Duplicate found. EEXIST");
				goto error_unlock;
			}
			break;
		
		case DICT_TYPE_ENUM:
			/* A type_enum object is linked in it's parent 'type' object lists 1 and 2 by its name and values */
			ret = insert_in_list ( &new->list[0], &_O(parent)->list[1], order_enum_by_name, (_dict_object_t **)ref );
			if (ret != 0) {
				TRACE_DEBUG(INFO, "Duplicate found. EEXIST");
				goto error_unlock;
			}
			ret = insert_in_list ( &new->list[1], &_O(parent)->list[2], order_enum_by_val, (_dict_object_t **)ref );
			if (ret != 0) {
				remove_from_list(&new->list[0]);
				TRACE_DEBUG(INFO, "Duplicate found. EEXIST");
				goto error_unlock;
			}
			break;
		
		case DICT_AVP:
			/* An avp object is linked in lists 1 and 2 of its vendor, by code and name */
			ret = insert_in_list ( &new->list[0], &vendor->list[1], order_avp_by_code, (_dict_object_t **)ref );
			if (ret != 0) {
				TRACE_DEBUG(INFO, "Duplicate found. EEXIST");
				goto error_unlock;
			}
			ret = insert_in_list ( &new->list[1], &vendor->list[2], order_avp_by_name, (_dict_object_t **)ref );
			if (ret != 0) {
				remove_from_list(&new->list[0]);
				TRACE_DEBUG(INFO, "Duplicate found. EEXIST");
				goto error_unlock;
			}
			break;
			
		case DICT_COMMAND:
			/* A command object is linked in g_list_cmd_name and g_list_cmd_code by its name and code */
			ret = insert_in_list ( &new->list[1], &g_list_cmd_code, order_cmd_by_codefl, (_dict_object_t **)ref );
			if (ret != 0) {
				TRACE_DEBUG(INFO, "Duplicate found. EEXIST");
				goto error_unlock;
			}
			ret = insert_in_list ( &new->list[0], &g_list_cmd_name, order_cmd_by_name, (_dict_object_t **)ref );
			if (ret != 0) {
				remove_from_list(&new->list[1]);
				TRACE_DEBUG(INFO, "Duplicate found. EEXIST");
				goto error_unlock;
			}
			break;
		
		case DICT_RULE:
			/* A rule object is linked in list[2] of its parent command or AVP by the name of the AVP it refers */
			ret = insert_in_list ( &new->list[0], &_O(parent)->list[2], order_rule_by_avpn, (_dict_object_t **)ref );
			if (ret != 0) {
				TRACE_DEBUG(INFO, "Duplicate found. EEXIST");
				goto error_unlock;
			}
			break;
			
		default:
			assert(0);
	}
	
	/* A new object has been created, increment the global counter */
	_OBINFO(new).count++;
	
	/* Unlock the dictionary */
	ret = pthread_rwlock_unlock(&dict_lock);
	if (ret != 0) {
		log_error("Unable to free a lock: %s\n", strerror(ret));
		TRACE_DEBUG (INFO, "pthread_rwlock_unlock failed." );
	}
	
	/* Save the pointer to the new object */
	if (ref)
		*ref = new;
	
	return 0;
	
error_unlock:
	(void)pthread_rwlock_unlock(&dict_lock);
error_free:
	free(new);
	return ret;
}

/* Search an object in the dictionary */
int dict_search ( dict_object_type_t type, int criteria, void * what, dict_object_t **result )
{
	int ret = 0;
	
	/* Check param */
	if (!CHECK_TYPE(type)) {
		TRACE_DEBUG (FULL, "EINVAL." );
		return EINVAL;
	}
	
	/* Lock the dictionary for reading */
	ret = pthread_rwlock_rdlock(&dict_lock);
	if (ret != 0) {
		log_error("Unable to acquire a lock: %s\n", strerror(ret));
		TRACE_DEBUG (INFO, "pthread_rwlock_rdlock failed." );
		return ret;
	}
	
	/* Now call the type-specific search function */
	ret = dict_obj_info[type].search_fct (criteria, what, (_dict_object_t **)result);
	
	/* Don't check return value here, but why would it fail... */
	(void)pthread_rwlock_unlock(&dict_lock);
	
	return ret;
}


/*******************************************************************************************************/
/*******************************************************************************************************/
/*                                                                                                     */
/*                                  The other module functions                                         */
/*                                                                                                     */
/*******************************************************************************************************/
/*******************************************************************************************************/


/* Initialize the dictionary */
int dict_init ( void )
{
	int ret = 0;
	
	/* Sanity checks */
	assert( (sizeof(type_base_name) / sizeof(type_base_name[0])) == (AVP_TYPE_MAX + 1) );
	assert( (sizeof(dict_obj_info)  / sizeof(dict_obj_info[0]))  == (DICT_TYPE_MAX + 1) );
	
	/* Initialize the lock for the dictionary */
	ret = pthread_rwlock_init(&dict_lock, NULL);
	if (ret != 0 ) {
		TRACE_DEBUG (INFO, "pthread_rwlock_init returned an error!");
		log_error("Initialization error: %s\n", strerror(ret));
		return ret;
	}
	
	/* Initialize the sentinel for vendors and AVP lists */
	init_object( &g_dict_vendors, DICT_VENDOR );
	g_dict_vendors.data.vendor.vendor_name = "(no vendor)";
	g_dict_vendors.list[0].o = NULL; /* overwrite since since element is also sentinel for this list. */
	
	
	/* Initialize the sentinel for applciations */
	init_object( &g_dict_applications, DICT_APPLICATION );
	g_dict_applications.data.application.application_name = "Diameter Common Messages";
	g_dict_applications.list[0].o = NULL; /* overwrite since since element is also sentinel for this list. */
			
	/* Initialize the sentinel for types */
	init_list ( &g_list_types, NULL );
	
	/* Initialize the sentinels for commands */
	init_list ( &g_list_cmd_name, NULL );
	init_list ( &g_list_cmd_code, NULL );
	
	/* Initialize the error command object */
	init_object( &g_dict_cmd_error, DICT_COMMAND );
	g_dict_cmd_error.data.cmd.cmd_name="(generic error format)";
	g_dict_cmd_error.data.cmd.cmd_flag_mask=CMD_FLAG_ERROR | CMD_FLAG_REQUEST | CMD_FLAG_RETRANSMIT;
	g_dict_cmd_error.data.cmd.cmd_flag_val =CMD_FLAG_ERROR;
	dict_cmd_error = (dict_object_t *) &g_dict_cmd_error;
	
	/* Initialize the objects defined in the Diameter Base Protocol */
	ret = dict_base_protocol();
	if (ret != 0) {
		TRACE_DEBUG (INFO, "dict_base_protocol() returned an error!");
		log_error("Initialization error: %s\n", strerror(ret));
		return ret;
	}
		
	/* Done */
	TRACE_DEBUG (INFO, "Dictionary module initialized");
	return 0;
}

/* Destroy the dictionary */
int dict_fini ( void )
{
	int ret = 0, i;
	
	TRACE_DEBUG (FULL, "Unloading dictionary...");
	
	/* Acquire the write lock to make sure no other operation is ongoing */
	ret = pthread_rwlock_wrlock(&dict_lock);
	if (ret != 0) {
		log_error("Unable to acquire a lock: %s\n", strerror(ret));
		TRACE_DEBUG (INFO, "pthread_rwlock_wrlock failed." );
		return ret;
	}
	
	/* Empty all the lists, free the elements */
	destroy_list ( &g_dict_cmd_error.list[2] );
	destroy_list ( &g_list_cmd_code );
	destroy_list ( &g_list_cmd_name );
	destroy_list ( &g_list_types );
	for (i=0; i< NB_LISTS_PER_OBJ; i++) {
		destroy_list ( &g_dict_applications.list[i] );
		destroy_list ( &g_dict_vendors.list[i] );
	}
	
	/* Dictionary is empty, now destroy the lock */
			
	ret = pthread_rwlock_unlock(&dict_lock);
	if (ret != 0 ) {
		TRACE_DEBUG (INFO, "pthread_rwlock_unlock returned an error!");
	}
	
	ret = pthread_rwlock_destroy(&dict_lock);
	if (ret != 0 ) {
		TRACE_DEBUG (INFO, "pthread_rwlock_destroy returned an error!");
	}
	
	return 0;
}

/* Iterate a callback on the rules for an object */
int dict_iterate_rules ( dict_object_t *parent, void * data, int (*cb)(void *, dict_rule_data_t *) )
{
	int ret = 0;
	_dict_list_t * li;
	
	/* Check parameters */
	if (!parent || verify_object(_O(parent))) {
		TRACE_DEBUG (FULL, "invalid parent object EINVAL." );
		return EINVAL;
	}
	if (  (_O(parent)->type != DICT_COMMAND) 
	   && ((_O(parent)->type != DICT_AVP) || (_O(parent)->data.avp.avp_basetype != AVP_TYPE_GROUPED) ) ) {
		TRACE_DEBUG (FULL, "wrong parent EINVAL." );
		return EINVAL;
	}
	TRACE_DEBUG (FULL, "Iterating on rules of %s: '%s'.", 
			_OBINFO(parent).name, 
			_O(parent)->type == DICT_COMMAND ? 
				  _O(parent)->data.cmd.cmd_name
				: _O(parent)->data.avp.avp_name);
	
	/* Acquire the read lock  */
	ret = pthread_rwlock_rdlock(&dict_lock);
	if (ret != 0) {
		log_error("Unable to acquire a lock: %s\n", strerror(ret));
		TRACE_DEBUG (INFO, "pthread_rwlock_rdlock failed." );
		return ret;
	}
	
	/* go through the list and call the cb on each rule data */
	for (li = &(_O(parent)->list[2]); li->next != &(_O(parent)->list[2]); li = li->next) {
		ret = (*cb)(data, &(li->next->o->data.rule));
		if (ret != 0)
			break;
	}
		
	/* Release the lock */
	{	
		int _ret = pthread_rwlock_unlock(&dict_lock);
		if (_ret != 0 ) {
			TRACE_DEBUG (INFO, "pthread_rwlock_unlock returned an error!");
			if (!ret)
				ret = _ret;
		}
	}
	return ret;
}	
